/*
 * 创建日期：2014年9月12日 下午0:0:02
 */
package com.cs.bd.commerce.util.io;

import android.content.ContentProvider;
import android.content.ContentValues;
import android.content.Context;
import android.content.SharedPreferences;
import android.database.Cursor;
import android.net.Uri;

import com.cs.bd.commerce.util.io.mp.IContentProvider;
import com.cs.bd.commerce.util.io.mp.ISharedPreferences;
import com.cs.bd.commerce.util.io.mp.MPSPImpl;

import java.util.Map;
import java.util.Set;

/**
 * 使用ContentProvider实现多进程SharedPreferences读写;<br>
 * 1、ContentProvider天生支持多进程访问；<br>
 * 2、使用内部私有BroadcastReceiver实现多进程OnSharedPreferenceChangeListener监听；<br>
 * 
 * 使用方法：AndroidManifest.xml中添加provider申明：<br>
 * <pre>
 * &lt;provider android:name="com.android.zgj.utils.MultiprocessSharedPreferences"
 * android:authorities="com.android.zgj.MultiprocessSharedPreferences"
 * android:process="com.android.zgj.MultiprocessSharedPreferences"
 * android:exported="false" /&gt;
 * &lt;!-- authorities属性里面最好使用包名做前缀，apk在安装时authorities同名的provider需要校验签名，否则无法安装；--!/&gt;<br>
 * </pre>
 * 
 * ContentProvider方式实现要注意：<br>
 * 1、当ContentProvider所在进程android.os.Process.killProcess(pid)时，会导致整个应用程序完全意外退出或者ContentProvider所在进程重启；<br>
 * 重启报错信息：Acquiring provider <processName> for user 0: existing object's process dead；<br>
 * 2、如果设备处在“安全模式”下，只有系统自带的ContentProvider才能被正常解析使用，因此put值时默认返回false，get值时默认返回null；<br>
 * 
 * 其他方式实现SharedPreferences的问题：<br>
 * 使用FileLock和FileObserver也可以实现多进程SharedPreferences读写，但是会有兼容性问题：<br>
 * 1、某些设备上卸载程序时锁文件无法删除导致卸载残留，进而导致无法重新安装该程序（报INSTALL_FAILED_UID_CHANGED错误）；<br>
 * 2、某些设备上FileLock会导致僵尸进程出现进而导致耗电；<br>
 * 3、僵尸进程出现后，正常进程的FileLock会一直阻塞等待僵尸进程中的FileLock释放，导致进程一直阻塞；<br>
 * 
 * added by matt
 * --------------------
 * 系统SharePreferences不能支持到多进程，原因如下：
 * After testing a lot with MODE_MULTI_PROCESS, I've three trials to share:
 * 1- Initialize the SharedPreferences once in each process and use it multiple times.
 *	  The problem: The values are not reflected in each process as expected. So each process has its own value of the SharedPreferences.
 * 2- Initialize the SharedPreferences in each put or get.
 *    This actually works and the value now is interchangeable between processes.
 *    The problem: sometimes after aggressively accessing the sharedpref, the shared preferences file got deleted with all its content, as described in this issue, and I get this warning in the log:
 *		W/FileUtils﹕ Failed to chmod(/data/data/com.hegazy.multiprocesssharedpref/shared_prefs/myprefs.xml): android.system.ErrnoException: chmod failed: ENOENT (No such file or directory)
 * 
 * 
 * added by matt==需要在AndroidManifest.xml里配置Provider
 * ---------------------
 * authorities配置规则：包名.commerce.MultiprocessSharedPreferences
 * android:process配置规则：所在进程不应调用android.os.Process.killProcess(pid) 
 *       <provider
 *           android:name="com.cs.bd.commerce.util.MultiprocessSharedPreferences"
 *           android:authorities="com.example.matt.test.commerce.MultiprocessSharedPreferences"
 *           android:exported="false"
 *           android:process=":test2" 
 *           />
 * 
 * @author zhangguojun
 * @version	1.0
 * @since JDK1.6
 * 
 * @来源 https://github.com/seven456/MultiprocessSharedPreferences
 */
public class MultiprocessSharedPreferences extends ContentProvider implements SharedPreferences {
	private IContentProvider mCPImpl;
	private ISharedPreferences mSPImpl;

	private IContentProvider getCPImpl() {
		if (null == mCPImpl) {
			synchronized (MultiprocessSharedPreferences.class) {
				if (null == mCPImpl) {
					mCPImpl = new MPSPImpl(getContext());
				}
			}
		}
		return mCPImpl;
	}

	/**
	 * mode不使用{@link Context#MODE_MULTI_PROCESS}特可以支持多进程了；
	 *
	 * @param mode
	 *
	 * @see Context#MODE_PRIVATE
	 * @see Context#MODE_WORLD_READABLE
	 * @see Context#MODE_WORLD_WRITEABLE
	 */
	public static SharedPreferences getSharedPreferences(Context context, String name, int mode) {
		return new MultiprocessSharedPreferences(context, name, mode);
	}

	/**
	 * @deprecated 此默认构造函数只用于父类ContentProvider在初始化时使用；
	 */
	@Deprecated
	public MultiprocessSharedPreferences() {

	}

	private MultiprocessSharedPreferences(Context context, String name, int mode) {
		mSPImpl = new MPSPImpl(context, name, mode, this);
	}


	@SuppressWarnings("unchecked")
	@Override
	public Map<String, ?> getAll() {
		return mSPImpl.getAll();
	}

	@Override
	public String getString(String key, String defValue) {
		return mSPImpl.getString(key, defValue);
	}

	// @Override // Android 3.0
	@SuppressWarnings("unchecked")
	public Set<String> getStringSet(String key, Set<String> defValues) {
		return mSPImpl.getStringSet(key, defValues);
	}

	@Override
	public int getInt(String key, int defValue) {
		return mSPImpl.getInt(key, defValue);
	}

	@Override
	public long getLong(String key, long defValue) {
		return mSPImpl.getLong(key, defValue);
	}

	@Override
	public float getFloat(String key, float defValue) {
		return mSPImpl.getFloat(key, defValue);
	}

	@Override
	public boolean getBoolean(String key, boolean defValue) {
		return mSPImpl.getBoolean(key, defValue);
	}

	@Override
	public boolean contains(String key) {
		return mSPImpl.contains(key);
	}

	@Override
	public Editor edit() {
		return mSPImpl.edit();
	}

	@Override
	public void registerOnSharedPreferenceChangeListener(OnSharedPreferenceChangeListener listener) {
		mSPImpl.registerOnSharedPreferenceChangeListener(listener);
	}

	@Override
	public void unregisterOnSharedPreferenceChangeListener(OnSharedPreferenceChangeListener listener) {
		mSPImpl.unregisterOnSharedPreferenceChangeListener(listener);
	}

	@Override
	public boolean onCreate() {
		getCPImpl().onCreate();
		return true;
	}

	@Override
	public Cursor query(Uri uri, String[] projection, String selection, String[] selectionArgs, String sortOrder) {
		return getCPImpl().query(uri, projection, selection, selectionArgs, sortOrder);
	}

	@Override
	public String getType(Uri uri) {
		return getCPImpl().getType(uri);
	}

	@Override
	public Uri insert(Uri uri, ContentValues values) {
		return getCPImpl().insert(uri, values);
	}

	@Override
	public int delete(Uri uri, String selection, String[] selectionArgs) {
		return getCPImpl().delete(uri, selection, selectionArgs);
	}

	@SuppressWarnings("unchecked")
	@Override
	public int update(Uri uri, ContentValues values, String selection, String[] selectionArgs) {
		return getCPImpl().update(uri, values, selection, selectionArgs);
	}
}