package com.cs.bd.commerce.util.retrofit;

import android.content.Context;
import android.content.SharedPreferences;

import com.cs.bd.commerce.util.CustomAlarm;
import com.cs.bd.commerce.util.retrofit.Interceptor.RepeatRequestCtrl;
import com.cs.bd.commerce.util.retrofit.Interceptor.RetryAfterNetOkCtrl;
import com.cs.bd.commerce.util.retrofit.cache.CustomCacheInterceptor;
import com.cs.bd.commerce.util.retrofit.cache.CustomCacheInterceptor.CacheType;

import java.io.IOException;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.TimeUnit;

import okhttp3.CacheControl;
import okhttp3.MCache;
import okhttp3.OkHttpClient;
import okhttp3.RequestBody;
import okhttp3.ResponseBody;
import retrofit2.Call;
import retrofit2.Callback;
import retrofit2.Response;
import retrofit2.Retrofit;

import static com.cs.bd.commerce.util.retrofit.Interceptor.LogInterceptor.LOG_SWITCH_KEY;
import static com.cs.bd.commerce.util.retrofit.RetrofitRequest.Method.get;
import static com.cs.bd.commerce.util.retrofit.RetrofitRequest.Method.post;
import static com.cs.bd.commerce.util.retrofit.RetrofitRequest.Method.put;

/**
 * Created by matt on 2017/11/2.
 */
public class RetrofitRequest {
    public static enum Method {
        get,
        post,
        put,
        delete,
    }

    String mUrl;
    Method mMethod;
    Map<String, String> mHeaderMap;
    /**
     * get请求参数
     */
    Map<String, String> mQueryMap;
    /**
     * post数据键值对
     */
    Map<String, String> mFieldMap;
    /**
     * post数据，优先级高于 {@link #mFieldMap}
     */
    RequestBody mRequestBody;
    /**
     * retrofit callback，优先级高于 {@link #mHttpCallback}
     */
    Callback mCallback;
    /**
     * 自封装的callback
     */
    RetrofitProxy.HttpCallback mHttpCallback;
    /**
     * 缓存设置
     */
    RequestCache mRequestCache;
    /**
     * 失败等网络好后重连设置
     */
    RequestRetryAfterNetOk mRequestRetryAfterNetOk;
    /**
     * 定时请求设置
     */
    RequestRepeat mRequestRepeat;
    /**
     * 定时请求业务类型
     */
    RepeatType mRepeatType = RepeatType.repeat_noraml;
    /**
     * 是否开启log
     */
    boolean mIsLog = false;

    /**
     * 正在执行的call(由于定时器和组合策略，可能一个request会有多个call)
     */
    private Call mCall;
    /**
     * 关联请求
     */
    private RetrofitRequest mAssociatedRequest;

    public RetrofitRequest(String url, Method method) {
        mUrl = url;
        mMethod = method;
    }

    public RetrofitRequest newBuilder() {
        RetrofitRequest clone = new RetrofitRequest(mUrl, mMethod);
        clone.mHeaderMap = mHeaderMap;
        clone.mQueryMap = mQueryMap;
        clone.mFieldMap = mFieldMap;
        clone.mRequestBody = mRequestBody;
        clone.mCallback = mCallback;
        clone.mHttpCallback = mHttpCallback;
        clone.mRequestCache = mRequestCache;
        clone.mRequestRetryAfterNetOk = mRequestRetryAfterNetOk;
        clone.mRequestRepeat = mRequestRepeat;
        clone.mRepeatType = mRepeatType;

        return clone;
    }

    private void checkValid() {
        //检查Request是否有效
        if (mRepeatType != RepeatType.repeat_noraml && mRequestRepeat == null) {
            throw new IllegalArgumentException("mRepeatType被设置后，必须同时设置mRequestRepeat!");
        }
    }

    /**
     * 设置Retrofit原生callback
     *
     * @param callback
     * @return
     */
    public RetrofitRequest setCallback(Callback<ResponseBody> callback) {
        mCallback = callback;
        return this;
    }

    /**
     * 设置封装callback
     *
     * @param httpCallback
     * @return
     */
    public RetrofitRequest setCallback(RetrofitProxy.HttpCallback<ResponseBody> httpCallback) {
        mHttpCallback = httpCallback;
        return this;
    }

    /**
     * 添加请求头
     *
     * @param key
     * @param value
     * @return
     */
    public RetrofitRequest addHeader(String key, String value) {
        if (mHeaderMap == null) {
            mHeaderMap = new HashMap<>();
        }
        mHeaderMap.put(key, value);
        return this;
    }

    /**
     * 设置请求头map
     *
     * @param headerMap
     * @return
     */
    public RetrofitRequest setHeaders(Map<String, String> headerMap) {
        mHeaderMap = headerMap;
        return this;
    }

    /**
     * 添加get请求参数
     *
     * @param key
     * @param value
     * @return
     */
    public RetrofitRequest addQuery(String key, String value) {
        if (mQueryMap == null) {
            mQueryMap = new HashMap<>();
        }
        mQueryMap.put(key, value);
        return this;
    }

    /**
     * 设置get请求参数map
     *
     * @param queryMap
     * @return
     */
    public RetrofitRequest setQueryMap(Map<String, String> queryMap) {
        mQueryMap = queryMap;
        return this;
    }

    /**
     * 添加post键值对
     *
     * @param key
     * @param value
     * @return
     */
    public RetrofitRequest addField(String key, String value) {
        if (mFieldMap == null) {
            mFieldMap = new HashMap<>();
        }
        mFieldMap.put(key, value);
        return this;
    }

    /**
     * 设置post键值对
     *
     * @param fieldMap
     * @return
     */
    public RetrofitRequest setFieldMap(Map<String, String> fieldMap) {
        mFieldMap = fieldMap;
        return this;
    }

    /**
     * 设置post body
     *
     * @param body
     * @return
     */
    public RetrofitRequest setRequestBody(RequestBody body) {
        mRequestBody = body;
        return this;
    }

    /**
     * 设置缓存
     *
     * @param requestCache
     * @return
     */
    public RetrofitRequest setRequestCache(RequestCache requestCache) {
        mRequestCache = requestCache;
        return this;
    }

    /**
     * 设置失败等网络好后自动重连
     *
     * @param requestRetryAfterNetOk
     * @return
     */
    public RetrofitRequest setRequestRetryAfterNetOk(RequestRetryAfterNetOk requestRetryAfterNetOk) {
        mRequestRetryAfterNetOk = requestRetryAfterNetOk;
        return this;
    }

    /**
     * 设置定时请求
     *
     * @param requestRepeat
     * @return
     */
    public RetrofitRequest setRequestRepeat(RequestRepeat requestRepeat) {
        mRequestRepeat = requestRepeat;
        return this;
    }

    /**
     * 设置http请求打印的log开关
     * @param log
     */
    public RetrofitRequest setLog(boolean log) {
        mIsLog = log;
        return this;
    }

    /**
     * 取消定时请求
     * instead of {@link #cancel(Context)}
     */
    @Deprecated
    public void cancelRepeatCall() {
        if (mRequestRepeat != null) {
            RepeatRequestCtrl.getInstance().cancelRepeatCall(mRequestRepeat.mRequestKey);
        }
    }

    /**
     * 设置定时请求的业务类型
     * @param repeatType
     */
    public RetrofitRequest setRepeatType(RepeatType repeatType) {
        mRepeatType = repeatType;
        return this;
    }

    /**
     * 异步执行
     *
     * @param context
     */
    public void enqueue(final Context context) {
        checkValid();

        if (mRequestRepeat != null && mRepeatType == RepeatType.cache_first_then_repeat) {
            enqueueCacheFirstThenRepeat(context);
            return;
        } else if (mRequestRepeat != null && mRepeatType == RepeatType.force_network_repeat) {
            enqueueForceNetworkRepeat(context);
            return;
        }

        if (mRequestRepeat == null && mRequestRetryAfterNetOk == null) {
            //无定时请求，无网络好后失败重试
            enqueueRaw(context);
        } else if (mRequestRepeat == null && mRequestRetryAfterNetOk != null) {
            //无定时请求，有网络好后失败重试
            enqueueRetryAfterNetOk(context);
        } else if (mRequestRepeat != null && mRequestRetryAfterNetOk == null) {
            //有定时请求，无网络好后失败重试
            enqueueRepeat(context, new Runnable() {
                @Override
                public void run() {
                    enqueueRaw(context);
                }
            });
        } else { //mRequestRepeat != null && mRequestRetryAfterNetOk != null
            //有定时请求，也有网络好后失败重试
            enqueueRepeat(context, new Runnable() {
                @Override
                public void run() {
                    enqueueRetryAfterNetOk(context);
                }
            });
        }
    }

    void enqueueCacheFirstThenRepeat(final Context context) {
        /**
         * 避免无限嵌套调用{@link #enqueue(Context)}
         */
        setRepeatType(RepeatType.repeat_noraml);

        final String requestKey = mRequestRepeat.mRequestKey;
        final Callback<ResponseBody> callback = mCallback != null ? mCallback : new RetrofitProxy.RetrofitCallback(mHttpCallback);

        setRequestRetryAfterNetOk(new RequestRetryAfterNetOk(requestKey, 1));
        mAssociatedRequest = newBuilder()
                .setRequestCache(RequestCache.forceCache(requestKey))
                .setRequestRepeat(null)
                .setCallback((RetrofitProxy.HttpCallback) null)
                .setCallback(new Callback<ResponseBody>() {
                    @Override
                    public void onResponse(Call<ResponseBody> call, Response<ResponseBody> response) {
                        if (response.raw().networkResponse() != null) {
                            mRequestRepeat.mAlarm.saveTriggerTime(mRequestRepeat.mSp, mRequestRepeat.mAlarmId);
                        }
                        callback.onResponse(call, response);

                        setRequestCache(RequestCache.forceNetworkOnly(requestKey)).enqueue(context);
                    }

                    @Override
                    public void onFailure(Call<ResponseBody> call, Throwable t) {
                        callback.onFailure(call, t);
                        if (!call.isCanceled() && !RetryAfterNetOkCtrl.getInstance(context).isRetry(requestKey)) {
                            enqueue(context);
                        }
                    }
                });
        mAssociatedRequest.enqueue(context);
    }

    void enqueueForceNetworkRepeat(final Context context) {
        /**
         * 避免无限嵌套调用{@link #enqueue(Context)}
         */
        setRepeatType(RepeatType.repeat_noraml);

        String requestKey = mRequestRepeat.mRequestKey;
        setRequestRetryAfterNetOk(new RequestRetryAfterNetOk(requestKey, 1));
        setRequestCache(RequestCache.forceNetworkOnly(requestKey)).enqueue(context);
    }

    void enqueueRaw(Context context) {
        mCall = buildCall(context);
        if (mCallback != null) {
            mCall.enqueue(mCallback);
        } else if (mHttpCallback != null) {
            mCall.enqueue(new RetrofitProxy.RetrofitCallback(mHttpCallback));
        } else {
            throw new IllegalArgumentException("执行enqueue方法，callback不能为空！");
        }
    }

    void enqueueRetryAfterNetOk(final Context context) {
        String requestKey = mRequestRetryAfterNetOk.mRequestKey;
        int maxRetryCount = mRequestRetryAfterNetOk.mMaxRetryCount;

        addHeader(RetryAfterNetOkCtrl.RETRY_AFTER_NET_OK_KEY, requestKey);
        //发起可失败重试的请求
        RetryAfterNetOkCtrl.getInstance(context).call(requestKey, maxRetryCount, new Runnable() {
            @Override
            public void run() {
                enqueueRaw(context);
            }
        });
    }

    void enqueueRepeat(Context context, Runnable r) {
        SharedPreferences sp = mRequestRepeat.mSp;
        CustomAlarm alarm = mRequestRepeat.mAlarm;
        int alarmId = mRequestRepeat.mAlarmId;
        long delay = mRequestRepeat.mDelay;
        long interval = mRequestRepeat.mInterval;
        String requestKey = mRequestRepeat.mRequestKey;

        addHeader(RepeatRequestCtrl.REPEAT_REQUEST_KEY, requestKey);
        //发起定时请求
        RepeatRequestCtrl.getInstance().callRepeat(sp, alarm, alarmId, delay, interval, requestKey, r);
    }

    /**
     * 同步执行
     *
     * @param context
     * @return
     * @throws IOException
     */
    public Response execute(Context context) throws IOException {
        checkValid();

        if (mRequestRetryAfterNetOk != null) {
            throw new IllegalArgumentException("execute方法不支持RequestFailRetry");
        }
        if (mRequestRepeat != null) {
            throw new IllegalArgumentException("execute方法不支持RequestRepeat");
        }
        mCall = buildCall(context);
        return mCall.execute();
    }

    /**
     * 取消请求
     */
    public void cancel(Context context) {
        //取消正在执行的请求
        if (mCall != null) {
            mCall.cancel();
        }
        //取消定时请求
        if (mRequestRepeat != null) {
            RepeatRequestCtrl.getInstance().cancelRepeatCall(mRequestRepeat.mRequestKey);
        }
        //取消失败重试
        if (mRequestRetryAfterNetOk != null) {
            RetryAfterNetOkCtrl.getInstance(context).cancel(mRequestRetryAfterNetOk.mRequestKey);
        }
        //取消关联请求
        if (mAssociatedRequest != null) {
            mAssociatedRequest.cancel(context);
        }
    }

    Call buildCall(Context context) {
        if (mIsLog) {
            addHeader(LOG_SWITCH_KEY, mUrl);
        }

        OkHttpClient client = ClientProvider.getInstance(context).get();
        if (mRequestCache != null) {
            client = client.newBuilder()
                    .addInterceptor(new CustomCacheInterceptor.Builder(
                            context, mRequestCache.mCacheType)
                            .setCacheControl(mRequestCache.mCacheControl)
                            .setCacheKey(mRequestCache.mCacheKey)
                            .setCache(mRequestCache.mCache)
                            .build()
                    )
                    .build();
        }
        Retrofit retrofit = new Retrofit.Builder()
                .client(client)
                .baseUrl(mUrl.substring(0, mUrl.lastIndexOf("/") + 1))
                .build();

        if (mHeaderMap == null) {
            mHeaderMap = new HashMap<>();
        }
        if (mMethod == get) {
            if (mQueryMap == null) {
                mQueryMap = new HashMap<>();
            }
            return retrofit.create(CommonService.class).get(mUrl, mHeaderMap, mQueryMap);
        } else if (mMethod == post) {
            if (mRequestBody != null) {
                return retrofit.create(CommonService.class).post(mUrl, mHeaderMap, mRequestBody);
            } else {
                if (mFieldMap == null) {
                    mFieldMap = new HashMap<>();
                }
                return retrofit.create(CommonService.class).post(mUrl, mHeaderMap, mFieldMap);
            }
        } else if (mMethod == put) {
            if (mRequestBody != null) {
                return retrofit.create(CommonService.class).put(mUrl, mHeaderMap, mRequestBody);
            } else {
                if (mFieldMap == null) {
                    mFieldMap = new HashMap<>();
                }
                return retrofit.create(CommonService.class).put(mUrl, mHeaderMap, mFieldMap);
            }
        } else /*if (mMethod == delete)*/ {
            if (mRequestBody != null) {
                return retrofit.create(CommonService.class).delete(mUrl, mHeaderMap, mRequestBody);
            } else {
                if (mFieldMap == null) {
                    mFieldMap = new HashMap<>();
                }
                return retrofit.create(CommonService.class).delete(mUrl, mHeaderMap, mFieldMap);
            }
        }
    }

    /**
     * 缓存参数
     */
    public static class RequestCache {
        CacheType mCacheType;
        CacheControl mCacheControl;
        String mCacheKey;
        MCache mCache;

        public RequestCache(CacheType cacheType, CacheControl cacheControl, String cacheKey) {
            mCacheType = cacheType;
            mCacheControl = cacheControl;
            mCacheKey = cacheKey;
        }

        /**
         * 不设置的话，默认Cache使用 {@link MCache#getsInstance(Context)}
         *
         * @param cache
         * @return
         */
        public RequestCache setCache(MCache cache) {
            mCache = cache;
            return this;
        }

        /**
         * 强制走缓存，不考虑有效期。没有缓存再走网络.
         *
         * @param cacheKey
         * @return
         */
        public static RequestCache forceCache(String cacheKey) {
            return new RequestCache(CacheType.cache_period_of_validity,
                    new CacheControl.Builder().maxStale(Integer.MAX_VALUE, TimeUnit.SECONDS).build()
                    , cacheKey);
        }

        /**
         * 强制只走缓存，不考虑有效期。没有缓存也不走网络.
         *
         * @param cacheKey
         * @return
         */
        public static RequestCache forceCacheOnly(String cacheKey) {
            return new RequestCache(CacheType.cache_period_of_validity, CacheControl.FORCE_CACHE, cacheKey);
        }

        /**
         * 取数据时不走缓存只走网络，返回的数据会保存到缓存里
         *
         * @return
         */
        public static RequestCache forceNetworkOnly(String cacheKey) {
            return new RequestCache(CacheType.cache_only_net, null, cacheKey);
        }
    }

    /**
     * 网络好后重试参数
     */
    public static class RequestRetryAfterNetOk {
        String mRequestKey;
        int mMaxRetryCount;

        public RequestRetryAfterNetOk(String requestKey, int maxRetryCount) {
            mRequestKey = requestKey;
            mMaxRetryCount = maxRetryCount;
        }
    }

    /**
     * 定时请求参数
     */
    public static class RequestRepeat {
        SharedPreferences mSp;
        CustomAlarm mAlarm;
        int mAlarmId;
        long mDelay;
        long mInterval;
        String mRequestKey;

        public RequestRepeat(SharedPreferences sp, CustomAlarm alarm, int alarmId,
                             long delay, long interval, String requestKey) {
            mSp = sp;
            mAlarm = alarm;
            mAlarmId = alarmId;
            mDelay = delay;
            mInterval = interval;
            mRequestKey = requestKey;
        }
    }

    /**
     * 定时请求类型
     */
    public static enum RepeatType {
        repeat_noraml,
        /**
         * 是否增强版的定时缓存机制：先返回缓存，再启动定时缓存机制。
         * <br/>需要设置 {@link #mRequestRepeat}
         * <br/>此时，忽略了 {@link #mRequestCache}、{@link #mRequestRetryAfterNetOk}
         * <br/>对应业务场景：比如定时取abtest配置-->应用启动时，先从缓存拿配置，然后设置闹钟定时更新配置
         */
        cache_first_then_repeat,
        /**
         * 是否增强版的定时缓存机制：定时从网络取数据放到缓存里并返回。
         * <br/>相当于{@link #cache_first_then_repeat}去掉第一步的"先返回缓存"
         * <br/>需要设置 {@link #mRequestRepeat}
         * <br/>此时，忽略了 {@link #mRequestCache}、{@link #mRequestRetryAfterNetOk}
         * <br/>对应业务场景：比如定时取abtest配置-->应用启动时，先从缓存拿配置，然后设置闹钟定时更新配置
         */
        force_network_repeat,
    }
}
